#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin(ibaldin@renci.org)
from typing import List, Dict

import networkx_query as nxq

from fim.graph.abc_property_graph import ABCPropertyGraphConstants, PropertyGraphQueryException
from fim.graph.resources.abc_bqm import ABCBQMPropertyGraph
from fim.graph.networkx_property_graph import NetworkXPropertyGraph, NetworkXGraphImporter


class NetworkXAggregateBQM(ABCBQMPropertyGraph, NetworkXPropertyGraph):
    """
    NetworkX mplementation of a *Aggregate* Broker Query Model (BQM)
    which aggregates site advertisements into  a single Composite Node.
    Intended to work with
    """

    def __init__(self, *, graph_id: str, importer, logger=None):
        super().__init__(graph_id=graph_id, importer=importer, logger=logger)

    def get_all_composite_nodes(self) -> List[str]:
        """
        Get a list of nodes IDs in a BQM model
        :return:
        """
        return self.get_all_nodes_by_class(label=ABCPropertyGraphConstants.CLASS_CompositeNode)

    def get_all_network_links(self) -> List[str]:
        """
        Get a list of link node ids in a BQM model
        :return:
        """
        return self.get_all_nodes_by_class(label=ABCPropertyGraphConstants.CLASS_Link)

    def find_node_by_name(self, node_name: str, label: str) -> str:
        """
        Return a node id of a node with this name
        :param node_name:
        :param label: label/class of the node
        :return:
        """
        assert node_name is not None
        my_graph = self.storage.get_graph(self.graph_id)
        graph_nodes = list(nxq.search_nodes(my_graph,
                                            {'and': [
                                                {'eq': [ABCPropertyGraphConstants.GRAPH_ID, self.graph_id]},
                                                {'eq': [ABCPropertyGraphConstants.PROP_NAME, node_name]},
                                                {'eq': [ABCPropertyGraphConstants.PROP_CLASS, label]}
                                            ]}))
        if len(graph_nodes) == 0:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                              msg=f"Unable to find node with name {node_name} class {label}")
        if len(graph_nodes) > 1:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                              msg=f"Graph contains multiple nodes with name {node_name} class {label}")

        return my_graph.nodes[graph_nodes[0]][ABCPropertyGraphConstants.NODE_ID]

    def check_node_name(self, *, node_id: str, label: str, name: str) -> bool:
        assert node_id is not None
        assert name is not None
        assert label is not None

        graph_nodes = list(nxq.search_nodes(self.storage.get_graph(self.graph_id),
                                            {'and': [
                                                {'eq': [ABCPropertyGraphConstants.GRAPH_ID, self.graph_id]},
                                                {'eq': [ABCPropertyGraphConstants.PROP_NAME, name]},
                                                {'eq': [ABCPropertyGraphConstants.PROP_CLASS, label]},
                                                {'eq': [ABCPropertyGraphConstants.NODE_ID, node_id]}
                                            ]}))
        return len(graph_nodes) > 0

    def get_all_node_or_component_connection_points(self, parent_node_id: str) -> List[str]:
        """
        Get a list of interfaces attached via switch fabrics to a composite node
        :param parent_node_id:
        :return:
        """
        assert parent_node_id is not None
        labels, parent_props = self.get_node_properties(node_id=parent_node_id)
        if ABCPropertyGraphConstants.CLASS_CompositeNode not in labels:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=parent_node_id,
                                              msg="Parent node type is not CompositeNode")
        sfs_ifs = self.get_first_and_second_neighbor(node_id=parent_node_id, rel1=ABCPropertyGraphConstants.REL_HAS,
                                                     node1_label=ABCPropertyGraphConstants.CLASS_SwitchFabric,
                                                     rel2=ABCPropertyGraphConstants.REL_CONNECTS,
                                                     node2_label=ABCPropertyGraphConstants.CLASS_ConnectionPoint)
        ret = list()
        # return only interface IDs, not interested in SwitchFabrics
        for tup in sfs_ifs:
            ret.append(tup[1])
        return ret


class NetworkXABQMFactory:
    """
    Help convert graphs between formats so long as they are rooted in NetworkXPropertyGraph
    """
    @staticmethod
    def create(graph: NetworkXPropertyGraph) -> NetworkXAggregateBQM:
        assert graph is not None
        assert isinstance(graph.importer, NetworkXGraphImporter)

        return NetworkXAggregateBQM(graph_id=graph.graph_id,
                                    importer=graph.importer,
                                    logger=graph.log)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
from typing import List
from abc import abstractmethod, ABC

from fim.graph.abc_property_graph import ABCPropertyGraph, PropertyGraphQueryException


class ABCBQMPropertyGraph(ABCPropertyGraph):
    """
    Very abstract BQM. All specialization is in subclasses that do
    whatever they need to do to turn a CBM into some kind of BQM
    """

    @abstractmethod
    def __init__(self, *, graph_id: str, importer, logger=None):
        super().__init__(graph_id=graph_id, importer=importer, logger=logger)

    def get_all_composite_node_components(self, parent_node_id: str) -> List[str]:
        """
        Return a list of components, children of a prent (presumably composite node)
        :param parent_node_id:
        :return:
        """
        assert parent_node_id is not None
        # check that parent is a NetworkNode
        labels, parent_props = self.get_node_properties(node_id=parent_node_id)
        if ABCPropertyGraph.CLASS_CompositeNode not in labels:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=parent_node_id,
                                              msg="Parent node type is not Composite")
        return self.get_first_neighbor(node_id=parent_node_id, rel=ABCPropertyGraph.REL_HAS,
                                       node_label=ABCPropertyGraph.CLASS_Component)




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
"""
NetworkX implementation of ADM (Aggregate Delegation Model) functionality.
"""

import uuid

from ..networkx_property_graph import NetworkXPropertyGraph, NetworkXGraphImporter
from .abc_arm import ABCARMPropertyGraph


class NetworkXARMGraph(ABCARMPropertyGraph, NetworkXPropertyGraph):

    def __init__(self, *, graph: NetworkXPropertyGraph, logger=None):
        """
        Initialize NetworkX ARM - supply an implementation of a graph
        :param graph:
        """
        super().__init__(graph=graph, logger=logger)


class NetworkXARMFactory:
    """
    Help convert graphs between formats so long as they are rooted in NetworkXPropertyGraph
    """
    @staticmethod
    def create(graph: NetworkXPropertyGraph) -> NetworkXARMGraph:
        assert graph is not None
        assert isinstance(graph.importer, NetworkXGraphImporter)

        return NetworkXARMGraph(graph=graph,
                                logger=graph.log)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
from abc import ABC


class ABCPropertyGraphConstants(ABC):

    # When a property is present, but not field, Neo4j returns word 'None'
    # NetworkX just has python None
    NEO4j_NONE = "None"

    SINGLE_POOL_NAME = "_"
    FIELD_POOL = "pool"
    FIELD_POOL_ID = "pool_id"
    FIELD_CAPACITIES = "capacities"
    FIELD_LABELS = "labels"
    PROP_CAPACITY_DELEGATIONS = "CapacityDelegations"
    PROP_LABEL_DELEGATIONS = "LabelDelegations"
    PROP_CAPACITIES = "Capacities"
    PROP_LABELS = "Labels"
    PROP_LABEL_ALLOCATIONS = "LabelAllocations"
    PROP_CAPACITY_ALLOCATIONS = "CapacityAllocations"
    GRAPH_ID = 'GraphID'
    NODE_ID = 'NodeID'
    PROP_NAME = 'Name'
    PROP_CLASS = 'Class'
    PROP_TYPE = 'Type'
    PROP_MODEL = 'Model'
    PROP_LAYER = 'Layer'
    PROP_TECHNOLOGY = 'Technology'
    PROP_SITE = 'Site'
    PROP_IMAGE_REF = 'ImageRef'
    PROP_MGMT_IP = 'MgmtIp'
    PROP_ALLOCATION_CONSTRAINTS = 'AllocationConstraints'
    PROP_SERVICE_ENDPOINT = 'ServiceEndpoint'
    PROP_DETAILS = 'Details'
    PROP_RESERVATION_INFO = 'ReservationInfo'
    PROP_NODE_MAP = "NodeMap"
    PROP_STRUCTURAL_INFO = "StructuralInfo"
    PROP_STITCH_NODE = "StitchNode"
    # these properties get validated to be valid JSON objects whenever someone validates the graph
    JSON_PROPERTY_NAMES = [PROP_LABELS, PROP_CAPACITIES, PROP_LABEL_DELEGATIONS,
                           PROP_CAPACITY_DELEGATIONS, PROP_LABEL_ALLOCATIONS,
                           PROP_CAPACITY_ALLOCATIONS, PROP_RESERVATION_INFO, PROP_STRUCTURAL_INFO]
    # these properties cannot be unset
    NO_UNSET_PROPERTIES = [GRAPH_ID, NODE_ID, PROP_TYPE, PROP_CLASS]

    CLASS_NetworkNode = 'NetworkNode'
    CLASS_Component = 'Component'
    CLASS_SwitchFabric = 'SwitchFabric'
    CLASS_ConnectionPoint = 'ConnectionPoint'
    CLASS_Link = 'Link'
    CLASS_CompositeLink = 'CompositeLink'
    CLASS_CompositeNode = 'CompositeNode'
    CLASS_MeasurementPoint = 'MeasurementPoint'

    REL_HAS = 'has'
    REL_CONNECTS = 'connects'
    REL_DEPENDS = 'depends'
    REL_ADAPTS = 'adapts'
    REL_PEERS = 'peers'


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
"""
Abstract definition of ASM (Abstract Slice Model) functionality
"""

from typing import List, Tuple
from abc import ABCMeta, abstractmethod

import json

from fim.slivers.network_node import NodeSliver
from fim.slivers.attached_components import ComponentSliver
from fim.slivers.switch_fabric import SwitchFabricSliver
from fim.slivers.interface_info import InterfaceSliver
from fim.slivers.network_link import NetworkLinkSliver
from fim.graph.abc_property_graph import ABCPropertyGraph, PropertyGraphQueryException


class ABCASMPropertyGraph(ABCPropertyGraph, metaclass=ABCMeta):
    """
    Interface for ASM
    """

    @abstractmethod
    def __init__(self, *, graph_id=str, importer, logger=None):
        super().__init__(graph_id=graph_id, importer=importer, logger=logger)

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'get_all_network_nodes') and
                callable(subclass.get_all_network_nodes) or NotImplemented)

    @abstractmethod
    def check_node_unique(self, *, label: str, name: str) -> bool:
        """
        Check no other node of this class/label and name exists. Doesn't
        apply universally as e.g. Component names are only unique
        within node, interface names are only unique within SF.
        :param label:
        :param name:
        :return:
        """

    @abstractmethod
    def check_node_name(self, *, node_id: str, label: str, name: str) -> bool:
        """
        Check if a node with this ID of this class/label has this name
        :param node_id:
        :param label:
        :param name:
        :return:
        """

    @abstractmethod
    def find_node_by_name(self, *, node_name: str, label: str) -> str:
        """
        Get node id of node based on its name and class/label. Throw
        exception if multiple matches found.
        :param node_name:
        :param label: node label or class of the node
        :return:
        """

    def get_all_network_nodes(self) -> List[str]:

        return self.get_all_nodes_by_class(label=ABCASMPropertyGraph.CLASS_NetworkNode)

    def get_all_network_links(self) -> List[str]:

        return self.get_all_nodes_by_class(label=ABCASMPropertyGraph.CLASS_Link)

    def set_mapping(self, *, node_id: str, to_graph_id: str, to_node_id: str) -> None:
        """
        Create a mapping from a node to another graph, node
        :param node_id:
        :param to_graph_id:
        :param to_node_id:
        :return:
        """
        assert node_id is not None
        assert to_graph_id is not None
        assert to_node_id is not None

        # save tuple as JSON list of guids
        l = [to_graph_id, to_node_id]
        jsonl = json.dumps(l)
        self.update_node_property(node_id=node_id, prop_name=ABCASMPropertyGraph.PROP_NODE_MAP,
                                  prop_val=jsonl)

    def get_mapping(self, *, node_id: str) -> Tuple[str, str] or None:
        """
        Retrieve a mapping, if exists for this node to another graph, node
        :param node_id:
        :return: graph_id, node_id tuple or None
        """
        assert node_id is not None
        # retrieve tuple as a json 2-member list
        _, props = self.get_node_properties(node_id=node_id)
        jsonl = props.get(ABCASMPropertyGraph.PROP_NODE_MAP, None)
        if jsonl is not None:
            l = json.loads(jsonl)
            assert(len(l) == 2)
            return tuple(l)
        return None

    def find_node_by_name_as_child(self, *, node_name: str, label: str, rel: str, parent_node_id: str) -> str or None:
        """
        Get node id of node based on its name, class label, as child of a parent node via a specified relationship
        or None
        :param node_name: 
        :param label: 
        :param rel: 
        :param parent_node_id: 
        :return: 
        """
        assert node_name is not None
        assert label is not None
        assert rel is not None
        assert parent_node_id is not None

        neighbs = self.get_first_neighbor(node_id=parent_node_id, rel=rel, node_label=label)
        for n in neighbs:
            _, props = self.get_node_properties(node_id=n)
            if props.get(ABCPropertyGraph.PROP_NAME, None) is not None and \
                props[ABCPropertyGraph.PROP_NAME] == node_name:
                return n
        raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                          msg=f"Unable to find node with name {node_name} "
                                              f"class {label} as child of {parent_node_id}")

    def get_all_network_node_components(self, parent_node_id: str) -> List[str]:
        """
        Return a list of components, children of a prent (presumably network node)
        :param parent_node_id:
        :return:
        """
        assert parent_node_id is not None
        # check that parent is a NetworkNode
        labels, parent_props = self.get_node_properties(node_id=parent_node_id)
        if ABCPropertyGraph.CLASS_NetworkNode not in labels:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=parent_node_id,
                                              msg="Parent node type is not NetworkNode")
        return self.get_first_neighbor(node_id=parent_node_id, rel=ABCPropertyGraph.REL_HAS,
                                       node_label=ABCPropertyGraph.CLASS_Component)

    def get_all_network_node_or_component_sfs(self, parent_node_id: str) -> List[str]:
        assert parent_node_id is not None
        # check that parent is a NetworkNode or Component
        labels, parent_props = self.get_node_properties(node_id=parent_node_id)
        if ABCPropertyGraph.CLASS_NetworkNode not in labels and \
            ABCPropertyGraph.CLASS_Component not in labels:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=parent_node_id,
                                              msg="Parent node type is not NetworkNode or Component")
        return self.get_first_neighbor(node_id=parent_node_id, rel=ABCPropertyGraph.REL_HAS,
                                       node_label=ABCPropertyGraph.CLASS_SwitchFabric)

    def get_all_node_or_component_connection_points(self, parent_node_id: str) -> List[str]:
        """
        Get a list of interfaces attached via switch fabrics
        :param parent_node_id:
        :return:
        """
        assert parent_node_id is not None
        labels, parent_props = self.get_node_properties(node_id=parent_node_id)
        if ABCPropertyGraph.CLASS_NetworkNode not in labels and \
                ABCPropertyGraph.CLASS_Component not in labels:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=parent_node_id,
                                              msg="Parent node type is not NetworkNode or Component")
        sfs_ifs = self.get_first_and_second_neighbor(node_id=parent_node_id, rel1=ABCPropertyGraph.REL_HAS,
                                                     node1_label=ABCPropertyGraph.CLASS_SwitchFabric,
                                                     rel2=ABCPropertyGraph.REL_CONNECTS,
                                                     node2_label=ABCPropertyGraph.CLASS_ConnectionPoint)
        ret = list()
        # return only interface IDs, not interested in SwitchFabrics
        for tup in sfs_ifs:
            ret.append(tup[1])
        return ret

    def get_all_sf_connection_points(self, parent_node_id: str) -> List[str]:
        assert parent_node_id is not None
        # check that parent is a SwitchFabric
        labels, parent_props = self.get_node_properties(node_id=parent_node_id)
        if ABCPropertyGraph.CLASS_SwitchFabric not in labels:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=parent_node_id,
                                              msg="Parent node type is not SwitchFabric")
        return self.get_first_neighbor(node_id=parent_node_id, rel=ABCPropertyGraph.REL_CONNECTS,
                                       node_label=ABCPropertyGraph.CLASS_ConnectionPoint)

    def find_component_by_name(self, *, parent_node_id: str, component_name: str) -> str:

        component_id_list = self.get_all_network_node_components(parent_node_id=parent_node_id)
        for cid in component_id_list:
            _, cprops = self.get_node_properties(node_id=cid)
            if cprops[ABCPropertyGraph.PROP_NAME] == component_name:
                return cid
        raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                          msg=f"Unable to find component with name {component_name}")

    def find_sf_by_name(self, *, parent_node_id: str, sfname: str) -> str:

        sf_id_list = self.get_all_network_node_or_component_sfs(parent_node_id=parent_node_id)
        for cid in sf_id_list:
            _, cprops = self.get_node_properties(node_id=cid)
            if cprops[ABCPropertyGraph.PROP_NAME] == sfname:
                return cid
        raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                          msg=f"Unable to find SwitchFabric with name {sfname}")

    def find_connection_point_by_name(self, *, parent_node_id: str, iname: str) -> str:

        if_id_list = self.get_all_sf_connection_points(parent_node_id=parent_node_id)
        for cid in if_id_list:
            _, cprops = self.get_node_properties(node_id=cid)
            if cprops[ABCPropertyGraph.PROP_NAME] == iname:
                return cid
        raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                          msg=f"Unable to find ConnectionPoint with name {iname}")

    def remove_network_node_with_components_sfs_cps_and_links(self, node_id: str):
        """
        Remove a network node, all of components and their interfaces, parent interfaces
        and connected links (as appropriate)
        :param node_id:
        :return:
        """
        # the network node itself
        # its components and switch fabrics
        # component switch fabrics
        # switch fabric interfaces (if present)
        # their parent interfaces (if present)
        # the 'Link' objects connected to interfaces if only one or no other interface connects to them

        # check we are a network node
        labels, node_props = self.get_node_properties(node_id=node_id)
        if ABCPropertyGraph.CLASS_NetworkNode not in labels:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_id,
                                              msg="This node type is not NetworkNode")
        # get a list of components
        components = self.get_first_neighbor(node_id=node_id, rel=ABCPropertyGraph.REL_HAS,
                                             node_label=ABCPropertyGraph.CLASS_Component)
        for cid in components:
            self.remove_component_with_sfs_cps_and_links(node_id=cid)

        # get a list of switch fabrics, delete the node
        switch_fabrics = self.get_first_neighbor(node_id=node_id, rel=ABCPropertyGraph.REL_HAS,
                                                 node_label=ABCPropertyGraph.CLASS_SwitchFabric)
        self.delete_node(node_id=node_id)
        for cid in switch_fabrics:
            self.remove_sf_with_cps_and_links(node_id=cid)

    def remove_component_with_sfs_cps_and_links(self, node_id: str):
        """
        Remove a component of a network node with all attached elements (switch fabrics
        and interfaces). Parent interfaces and links are removed if they have no other
        children or connected interfaces.
        :param node_id: component node id
        :return:
        """
        # check we are a component
        labels, node_props = self.get_node_properties(node_id=node_id)
        if ABCPropertyGraph.CLASS_Component not in labels:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_id,
                                              msg="This node type is not Component")
        # get a list of SwitchFabrics, delete the node
        switch_fabrics = self.get_first_neighbor(node_id=node_id, rel=ABCPropertyGraph.REL_HAS,
                                                 node_label=ABCPropertyGraph.CLASS_SwitchFabric)
        self.delete_node(node_id=node_id)
        for sf in switch_fabrics:
            self.remove_sf_with_cps_and_links(node_id=sf)

    def remove_network_link(self, node_id: str):

        # check we are a link
        labels, node_props = self.get_node_properties(node_id=node_id)
        if ABCPropertyGraph.CLASS_Link not in labels:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_id,
                                              msg="This node type is not Link")
        # edges to interfaces/connection points will disappear
        self.delete_node(node_id=node_id)

    def remove_sf_with_cps_and_links(self, node_id: str):
        """
        Remove switch fabric with subtending interfaces and links
        :param node_id:
        :return:
        """
        # check we are a switch fabric
        labels, node_props = self.get_node_properties(node_id=node_id)
        if ABCPropertyGraph.CLASS_SwitchFabric not in labels:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_id,
                                              msg="This node type is not SwitchFabric")
        # get a list of interfaces, delete the switch fabric
        interfaces = self.get_first_neighbor(node_id=node_id, rel=ABCPropertyGraph.REL_CONNECTS,
                                             node_label=ABCPropertyGraph.CLASS_ConnectionPoint)
        self.delete_node(node_id=node_id)
        for iif in interfaces:
            self.remove_cp_and_links(node_id=iif)

    def remove_cp_and_links(self, node_id: str):
        """
        Remove ConnectionPoint and links. Parent ConnectionPoints and links are removed
        if they have no other children or other connected connection points.
        :param node_id: interface node id
        :return:
        """
        # some interfaces may have parent interfaces, which can be deleted if they only have the
        # one child
        interfaces_to_delete = {node_id}
        parents = self.get_first_neighbor(node_id=node_id,
                                          rel=ABCPropertyGraph.REL_CONNECTS,
                                          node_label=ABCPropertyGraph.CLASS_ConnectionPoint)
        for parent in parents:  # really should only be one parent interface
            children = self.get_first_neighbor(node_id=parent,
                                               rel=ABCPropertyGraph.REL_CONNECTS,
                                               node_label=ABCPropertyGraph.CLASS_ConnectionPoint)
            if len(children) == 1:  # if only child, can delete parent
                interfaces_to_delete.add(parent)
        # interfaces themselves and parent interfaces
        # may be connected to links which can be deleted if nothing
        # else connects to them
        links_to_delete = set()
        for interface in interfaces_to_delete:
            links = self.get_first_neighbor(node_id=interface,
                                            rel=ABCPropertyGraph.REL_CONNECTS,
                                            node_label=ABCPropertyGraph.CLASS_Link)
            for link in links:  # should only be one
                connected_interfaces = self.get_first_neighbor(node_id=link,
                                                               rel=ABCPropertyGraph.REL_CONNECTS,
                                                               node_label=ABCPropertyGraph.CLASS_ConnectionPoint)
                if len(connected_interfaces) == 2:  # connected to us and another thing, can delete
                    links_to_delete.add(link)
        # delete nodes in these sets
        for deleted_id in interfaces_to_delete.union(links_to_delete):
            self.delete_node(node_id=deleted_id)

    def add_network_node_sliver(self, *, sliver: NodeSliver):

        assert sliver is not None
        assert sliver.node_id is not None
        assert self.check_node_unique(label=ABCPropertyGraph.CLASS_NetworkNode, name=sliver.resource_name)

        props = self.node_sliver_to_graph_properties_dict(sliver)
        self.add_node(node_id=sliver.node_id, label=ABCPropertyGraph.CLASS_NetworkNode, props=props)
        # if components aren't empty, add components, their switch fabrics and interfaces
        aci = sliver.attached_components_info
        if aci is not None:
            for csliver in aci.devices.values():
                self.add_component_sliver(parent_node_id=sliver.node_id,
                                          component=csliver)
        # if switch fabrics arent empty add them with their interfaces
        sfi = sliver.switch_fabric_info
        if sfi is not None:
            for sf in sfi.switch_fabrics.values():
                self.add_switch_fabric_sliver(parent_node_id=sliver.node_id,
                                              switch_fabric=sf)

    def add_network_link_sliver(self, *, lsliver: NetworkLinkSliver, interfaces: List[str]):

        assert lsliver is not None
        assert lsliver.node_id is not None
        assert self.check_node_unique(label=ABCPropertyGraph.CLASS_Link, name=lsliver.resource_name)
        assert interfaces is not None

        props = self.link_sliver_to_graph_properties_dict(lsliver)
        self.add_node(node_id=lsliver.node_id, label=ABCPropertyGraph.CLASS_Link, props=props)
        # add edge links to specified interfaces
        for i in interfaces:
            self.add_link(node_a=lsliver.node_id, rel=ABCPropertyGraph.REL_CONNECTS, node_b=i)

    def add_component_sliver(self, *, parent_node_id: str, component: ComponentSliver):
        """
        Add network node component as a sliver, linking back to parent
        :param parent_node_id:
        :param component:
        :return:
        """
        assert component.node_id is not None
        assert parent_node_id is not None

        props = self.component_sliver_to_graph_properties_dict(component)
        self.add_node(node_id=component.node_id, label=ABCPropertyGraph.CLASS_Component, props=props)
        self.add_link(node_a=parent_node_id, rel=ABCPropertyGraph.REL_HAS, node_b=component.node_id)
        sfi = component.switch_fabric_info
        if sfi is not None:
            for sf in sfi.switch_fabrics.values():
                self.add_switch_fabric_sliver(parent_node_id=component.node_id,
                                              switch_fabric=sf)

    def add_switch_fabric_sliver(self, *, parent_node_id: str, switch_fabric: SwitchFabricSliver):
        """
        Add switch fabric as a sliver to either node or component, linking back to parent.
        :param parent_node_id:
        :param switch_fabric:
        :return:
        """
        assert parent_node_id is not None
        assert switch_fabric.node_id is not None

        props = self.switch_fabric_sliver_to_graph_properties_dict(switch_fabric)
        self.add_node(node_id=switch_fabric.node_id, label=ABCPropertyGraph.CLASS_SwitchFabric, props=props)
        self.add_link(node_a=parent_node_id, rel=ABCPropertyGraph.REL_HAS, node_b=switch_fabric.node_id)
        ii = switch_fabric.interface_info
        if ii is not None:
            for i in ii.interfaces.values():
                self.add_interface_sliver(parent_node_id=switch_fabric.node_id,
                                          interface=i)

    def add_interface_sliver(self, *, parent_node_id: str, interface: InterfaceSliver):
        """
        Add interface to a switch fabric, linking back to parent.
        :param parent_node_id: switch fabric id
        :param interface: interface sliver description
        :return:
        """
        assert interface.node_id is not None
        assert parent_node_id is not None

        props = self.interface_sliver_to_graph_properties_dict(interface)
        self.add_node(node_id=interface.node_id, label=ABCPropertyGraph.CLASS_ConnectionPoint, props=props)
        self.add_link(node_a=parent_node_id, rel=ABCPropertyGraph.REL_CONNECTS, node_b=interface.node_id)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
"""
Abstract definition of ADM (Aggregate Delegation Model) functionality
"""
import json

from collections import defaultdict

from abc import ABCMeta, abstractmethod
from ..abc_property_graph import ABCPropertyGraph, PropertyGraphQueryException
from ...slivers.delegations import Delegations, Pools, DelegationType


class ABCADMPropertyGraph(ABCPropertyGraph):
    """
    Interface for an ADM Mixin on top of a property graph
    """

    @abstractmethod
    def __init__(self, *, graph_id: str, importer, logger=None):
        super().__init__(graph_id=graph_id, importer=importer, logger=logger)

    def rewrite_delegations(self, *, real_adm_id: str = None) -> None:
        """
        Rewrite label and capacity delegations on all nodes to be dictionaries
        referenced by ADM graph id. Note that external code should
        not interact with delegations on ADM graphs.
        Sometimes ADMs are cloned into temporary graphs so the method provides
        a way to pass original ADM id as an option.
        :param real_adm_id:
        :return:
        """
        for node_id in self.list_all_node_ids():
            _, node_props = self.get_node_properties(node_id=node_id)
            props_modified = False
            for delegation_prop_name in [self.PROP_LABEL_DELEGATIONS, self.PROP_CAPACITY_DELEGATIONS]:
                if delegation_prop_name in node_props.keys():
                    if node_props.get(delegation_prop_name, None) is None or \
                            node_props[delegation_prop_name] == self.NEO4j_NONE:
                        continue
                    props_modified = True
                    # turn delegation into Python object
                    delegations = Delegations.from_json(json_str=node_props[delegation_prop_name],
                                                        atype=DelegationType.LABEL if
                                                        delegation_prop_name == self.PROP_LABEL_DELEGATIONS else
                                                        DelegationType.CAPACITY)
                    if len(delegations.get_delegation_ids()) != 1:
                        raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_id,
                                                          msg=f'This node contains delegation structure with more'
                                                              f'than one entry {delegations}')
                    # replace the one delegation id that should be present here with a graph id
                    # CAUTION: We directly manipulate the delegations object innards here for efficiency
                    del_id = list(delegations.get_delegation_ids())[0]
                    delegation = delegations.get_by_delegation_id(del_id)
                    if real_adm_id is None:
                        delegation.delegation_id = self.graph_id
                    else:
                        delegation.delegation_id = real_adm_id
                    delegations.delegations[delegation.delegation_id] = delegations.delegations.pop(del_id)
                    node_props[delegation_prop_name] = delegations.to_json()
            if props_modified:
                # write back
                self.update_node_properties(node_id=node_id, props=node_props)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
from typing import List

from fim.graph.abc_property_graph import PropertyGraphQueryException
from fim.graph.neo4j_property_graph import Neo4jPropertyGraph, Neo4jGraphImporter
from fim.graph.slices.abc_asm import ABCASMPropertyGraph


class Neo4jASM(ABCASMPropertyGraph, Neo4jPropertyGraph):

    def __init__(self, *, graph_id=str, importer, logger=None):
        super().__init__(graph_id=graph_id, importer=importer, logger=logger)

    def check_node_unique(self, *, label: str, name: str) -> bool:
        assert label is not None
        assert name is not None

        query = f"MATCH (n:{label} {{GraphID: $graphId, Name: $name}}) RETURN collect(n.NodeID) as nodeids"
        with self.driver.session() as session:
            val = session.run(query, graphId=self.graph_id, name=name).single()
            if val is None or len(val.data()) == 0 or len(val.data()['nodeids']) == 0:
                return True
            return False

    def check_node_name(self, *, node_id: str, label: str, name: str) -> bool:
        assert node_id is not None
        assert label is not None
        assert name is not None

        query = f"MATCH (n:{label} {{GraphID: $graphId, NodeID: $nodeId, Name: $name}}) " \
                f"RETURN collect(n.NodeID) as nodeids"
        with self.driver.session() as session:
            val = session.run(query, graphId=self.graph_id, nodeId=node_id, name=name).single()
            if val is None or len(val.data()) == 0:
                return False
            return len(val.data()['nodeids']) > 0

    def find_node_by_name(self, *, node_name: str, label: str) -> str:

        assert node_name is not None
        assert label is not None
        query = f"MATCH(n:{label} {{GraphID: $graphId, Name: $name }}) RETURN collect(n.NodeID) as nodeids"
        with self.driver.session() as session:
            val = session.run(query, graphId=self.graph_id, name=node_name).single()
            if val is None or len(val.data()) == 0:
                raise PropertyGraphQueryException(graph_id=self.graph_id,
                                                  node_id=None,
                                                  msg=f"Unable to find node with name {node_name} class {label}")
            if len(val.data()['nodeids']) > 1:
                    raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                                      msg=f"Graph contains multiple nodes "
                                                          f"with name {node_name} class {label}")
            return val.data()['nodeids'][0]


class Neo4jASMFactory:
    """
    Help convert graphs between formats so long as they are rooted in Neo4jPropertyGraph
    """
    @staticmethod
    def create(graph: Neo4jPropertyGraph) -> Neo4jASM:
        assert graph is not None
        assert isinstance(graph.importer, Neo4jGraphImporter)

        return Neo4jASM(graph_id=graph.graph_id,
                        importer=graph.importer,
                        logger=graph.log)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
"""
Neo4j implementation of ADM (Aggregate Delegation Model) functionality.
"""

import uuid

from ..neo4j_property_graph import Neo4jPropertyGraph, Neo4jGraphImporter
from .abc_adm import ABCADMPropertyGraph


class Neo4jADMGraph(ABCADMPropertyGraph, Neo4jPropertyGraph):

    def __init__(self, *, graph_id: str = None, importer: Neo4jGraphImporter, logger=None):
        """
        Initialize CBM either from existing ID or generate new graph id
        :param graph_id:
        :param importer:
        """
        if graph_id is None:
            graph_id = str(uuid.uuid4())

        super().__init__(graph_id=graph_id, importer=importer, logger=logger)


class Neo4jADMFactory:
    """
    Help convert graphs between formats so long as they are rooted in Neo4jPropertyGraph
    """
    @staticmethod
    def create(graph: Neo4jPropertyGraph) -> Neo4jADMGraph:
        assert graph is not None
        assert isinstance(graph.importer, Neo4jGraphImporter)

        return Neo4jADMGraph(graph_id=graph.graph_id,
                             importer=graph.importer,
                             logger=graph.log)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)

from typing import Dict, Any, List, Tuple

from fim.view_only_dict import ViewOnlyDict
from .node import Node
from .interface import Interface

from ..graph.abc_property_graph import ABCPropertyGraph


class CompositeNode(Node):
    """
    A composite node of the topology. In addition to public methods the following calls
    return various dictionaries or lists:
    node.components - a dictionary of components
    node.interfaces - a dictionary of all interfaces
    node.interface_list - a list of all interfaces
    """

    def __init__(self, *, name: str, node_id: str, topo: Any):
        """
        Don't call this method yourself, call topology.add_node()
        node_id will be generated if not provided for experiment topologies

        :param name:
        :param node_id:
        :param topo:
        """
        assert name is not None
        assert topo is not None
        assert node_id is not None

        # skip one level of constructor
        super(Node, self).__init__(name=name, node_id=node_id, topo=topo)
        # check that this node exists
        existing_node_id = self.topo.graph_model.find_node_by_name(node_name=name,
                                                                   label=str(ABCPropertyGraph.CLASS_CompositeNode))
        if existing_node_id != node_id:
            raise RuntimeError(f'Composite Node name {name} is not unique within the topology')

    def __list_components(self) -> ViewOnlyDict:
        """
        List all Components children of a node in the topology as a dictionary
        organized by component name. Modifying the dictionary will not affect
        the underlying model, but modifying Components in the dictionary will.
        :return:
        """
        node_id_list = self.topo.graph_model.get_all_composite_node_components(parent_node_id=self.node_id)
        ret = dict()
        for nid in node_id_list:
            c = self._get_component_by_id(nid)
            ret[c.name] = c
        return ViewOnlyDict(ret)

    def __get_interface_by_id(self, node_id: str) -> Interface:
        """
        Get an interface of a node by its node_id, return Interface object
        :param node_id:
        :return:
        """
        assert node_id is not None
        clazzes, node_props = self.topo.graph_model.get_node_properties(node_id=node_id)
        assert node_props.get(ABCPropertyGraph.PROP_NAME, None) is not None
        assert ABCPropertyGraph.CLASS_ConnectionPoint in clazzes
        return Interface(name=node_props[ABCPropertyGraph.PROP_NAME], node_id=node_id,
                         topo=self.topo)

    def __list_interfaces(self) -> ViewOnlyDict:
        """
        List all interfaces of composite node ignoring the components
        :return:
        """
        # immediately-attached interfaces only for composite nodes
        node_if_list = self.topo.graph_model.get_all_node_or_component_connection_points(parent_node_id=self.node_id)
        direct_interfaces = dict()
        for nid in node_if_list:
            i = self.__get_interface_by_id(nid)
            direct_interfaces[i.name] = i
        return ViewOnlyDict(direct_interfaces)

    def __list_of_interfaces(self) -> Tuple[Any]:
        """
        List all interfaces of node and its components
        :return:
        """
        return tuple(self.__list_interfaces().values())

    def __getattr__(self, item):
        """
        Special handling for attributes like 'components' and 'interfaces' -
        which query into the model. They return dicts and list
        containers. Modifying containers does not affect the underlying
        graph mode, but modifying elements of lists or values of dicts does.
        :param item:
        :return:
        """
        if item == 'components':
            return self.__list_components()
        if item == 'interfaces':
            return self.__list_interfaces()
        if item == 'interface_list':
            return self.__list_of_interfaces()
        raise RuntimeError(f'Attribute {item} not available')

    def __repr__(self):
        _, node_properties = self.topo.graph_model.get_node_properties(node_id=self.node_id)
        node_sliver = self.topo.graph_model.node_sliver_from_graph_properties_dict(node_properties)
        return node_sliver.__repr__()

    def __str__(self):
        return self.__repr__()





#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
"""
Label and capacity delegations and pools. Each delegation knows which node it is
defined on, each pool knows which node defined it and which nodes it
applies to and can be set/queried for that information.
"""

from enum import Enum
from typing import List, Dict, Set, Tuple
import json

from fim.graph.abc_property_graph_constants import ABCPropertyGraphConstants
from .capacities_labels import Capacities, Labels


class DelegationType(Enum):
    CAPACITY = 1
    LABEL = 2


class DelegationFormat(Enum):
    PoolDefinition = 1
    PoolReference = 2
    SinglePool = 3


class Delegation:
    """
    Single label and capacity delegation (pool definition or reference).
    Encodes a single entry in a dictionary describing multiple delegations.
    This object by itself is not JSON serializable, for that look at Delegations object.
    """
    def __init__(self, *, atype: DelegationType, delegation_id: str,
                 aformat: DelegationFormat = DelegationFormat.SinglePool,
                 pool_id: str = None):
        """
        Define a delegation of resources on a given node (for a given pool - either
        as definition or reference)
        :param atype:
        :param delegation_id:
        :param aformat:
        :param pool_id:
        """
        assert atype is not None
        assert delegation_id is not None
        self.type = atype
        self.format = aformat
        self.delegation_id = delegation_id
        self.delegation_details = None
        self.pool_id = pool_id
        if aformat != DelegationFormat.SinglePool:
            assert pool_id is not None

    def get_delegation_type(self) -> DelegationType:
        return self.type

    def get_pool_name(self) -> str:
        return self.pool_id

    def get_format(self) -> DelegationFormat:
        return self.format

    def get_delegation_id(self) -> str:
        return self.delegation_id

    def set_details(self, caporlab) -> None:
        """
        set details of the delegation dictionary from either a Labels or Capacities
        object
        :param caporlab: either Labels or Capacities
        :return:
        """
        assert caporlab is not None
        assert (isinstance(caporlab, Labels) or isinstance(caporlab, Capacities))
        if self.format == DelegationFormat.PoolReference:
            raise DelegationException(msg=f'Trying to add Labels or Capacities object to PoolReference delegation')
        if (isinstance(caporlab, Labels) and self.type == DelegationType.CAPACITY) or \
            (isinstance(caporlab, Capacities) and self.type == DelegationType.LABEL):
            raise DelegationException(msg=f'Trying to add Capacities to a LABEL type delegation or vice versa')
        self.delegation_details = caporlab

    def get_details(self) -> Labels or Capacities:
        """
        get delegation details as either Labels or Capacities object
        :return:
        """
        if self.delegation_details is None:
            return None
        return self.delegation_details

    def get_details_as_dict(self) -> Dict[str, str] or None:
        """
        get details of labels/capacities as a dictionary
        :return:
        """
        if self.delegation_details is None:
            return None
        return self.delegation_details.to_dict()

    def __repr__(self) -> str:
        return f"{self.type} delegation delegated to {self.delegation_id}: " \
               f"format {self.format} for pool {self.pool_id} with {self.delegation_details} "


class Delegations:
    """
    Multiple delegations referenced by delegation id within them on a single
    model element. Serializable to and from JSON so used in models and slivers.
    Delegation id can be graph id or a unique string (depending on whether this
    is on CBM, or ADM/ARM)
    """
    def __init__(self, *, atype: DelegationType):

        self.type = atype
        self.delegations = dict()

    def add_delegations(self, *args) -> None:
        """
        Include well-formed Delegation objects into the dictionary.
        Delegation types must match the type of this container object.
        :param args: sequence of Delegation objects
        :return:
        """
        for delegation in args:
            assert delegation.delegation_id is not None
            assert delegation.get_delegation_type() == self.type

            if self.delegations.get(delegation.delegation_id, None) is not None:
                raise DelegationException(msg=f'Delegation with id {delegation.delegation_id} is already present')
            if self.type != delegation.get_delegation_type():
                raise DelegationException(msg=f'Delegation with id {delegation.delegation_id} type '
                                              f'{delegation.get_delegation_type()} does not match delegations '
                                              f'type {self.type}')
            self.delegations[delegation.delegation_id] = delegation

    def get_by_delegation_id(self, delegation_id: str) -> Delegation:
        """
        retrieve a list of delegations by their id or None
        :param delegation_id:
        :return:
        """
        assert delegation_id is not None
        return self.delegations.get(delegation_id, None)

    def get_sole_delegation(self) -> Tuple[str, Delegation]:
        """
        Checks that delegations object has just one delegation and returns
        a tuple of delegation id and delegation object. Useful primarily on CBMs.
        :return:
        """
        if len(self.delegations) != 1:
            raise DelegationException(msg=f"Expected to find only one delegation in Delegations object {self}")
        for k, v in self.delegations.items():
            return k, v

    def get_delegations_as_list(self) -> List[Delegation]:
        """
        Return individual delegation objects as a list
        :return:
        """
        return list(self.delegations.values())

    def get_delegation_ids(self) -> Set:
        """
        get a set of all delegation ids
        :return:
        """
        return set(self.delegations.keys())

    def remove_by_id(self, delegation_id: str) -> None:
        """
        Remove a delegation with this id from container
        :param delegation_id:
        :return:
        """
        assert delegation_id is not None
        if delegation_id not in self.delegations.keys():
            return
        self.delegations.pop(delegation_id)

    def return_delegations_for_id(self, delegation_id: str):
        """
        Return a delegations object limited to this delegation id or None.
        :param delegation_id:
        :return:
        """
        if delegation_id not in self.delegations.keys():
            return None
        ds = Delegations(atype=self.type)
        ds.add_delegations(self.delegations[delegation_id])
        return ds

    def to_json(self) -> str:
        """
        convert object to JSON. Example objects:
        { "del1": {
            "pool_name": "_",
            "capacities": { <capacities dictionary> } }
        or
        { "del1": {
            "pool_name": "pool1",
            "labels": { <labels dictionary> } }
        or
        { "del1": {
            "pool": "pool1" }
        :return:
        """
        json_dict = {}
        for k, v in self.delegations.items():
            inner_dict = {}
            if v.get_format() == DelegationFormat.SinglePool:
                assert v.get_details_as_dict() is not None
                inner_dict[ABCPropertyGraphConstants.FIELD_POOL_ID] = ABCPropertyGraphConstants.SINGLE_POOL_NAME
                if self.type == DelegationType.CAPACITY:
                    inner_dict[ABCPropertyGraphConstants.FIELD_CAPACITIES] = v.get_details_as_dict()
                else:
                    inner_dict[ABCPropertyGraphConstants.FIELD_LABELS] = v.get_details_as_dict()
            elif v.get_format() == DelegationFormat.PoolDefinition:
                assert v.get_pool_name() is not None
                assert v.get_details_as_dict() is not None
                inner_dict[ABCPropertyGraphConstants.FIELD_POOL_ID] = v.get_pool_name()
                if self.type == DelegationType.CAPACITY:
                    inner_dict[ABCPropertyGraphConstants.FIELD_CAPACITIES] = v.get_details_as_dict()
                else:
                    inner_dict[ABCPropertyGraphConstants.FIELD_LABELS] = v.get_details_as_dict()
            elif v.get_format() == DelegationFormat.PoolReference:
                assert v.get_pool_name() is not None
                inner_dict[ABCPropertyGraphConstants.FIELD_POOL] = v.get_pool_name()
            json_dict[k] = inner_dict

        return json.dumps(json_dict)

    @classmethod
    def from_json(cls, *, json_str: str, atype: DelegationType):
        """
        convert from a JSON representation producing a Delegations object. Returns None
        if json_str is None
        :param json_str:
        :param atype: one of DelegationType
        :return:
        """
        if json_str is None or len(json_str) == 0 or json_str == ABCPropertyGraphConstants.NEO4j_NONE:
            return None
        ds = Delegations(atype=atype)
        json_dict = json.loads(json_str)
        for k, v in json_dict.items():
            if ABCPropertyGraphConstants.FIELD_POOL_ID in v.keys():
                # single element pool or pool definition
                if v[ABCPropertyGraphConstants.FIELD_POOL_ID] == ABCPropertyGraphConstants.SINGLE_POOL_NAME:
                    format = DelegationFormat.SinglePool
                    pool_id = None
                else:
                    format = DelegationFormat.PoolDefinition
                    pool_id = v[ABCPropertyGraphConstants.FIELD_POOL_ID]
                if atype == DelegationType.CAPACITY:
                    caporlabdict = v[ABCPropertyGraphConstants.FIELD_CAPACITIES]
                    caporlab = Capacities().set_fields(**caporlabdict)
                else:
                    caporlabdict = v[ABCPropertyGraphConstants.FIELD_LABELS]
                    caporlab = Labels().set_fields(**caporlabdict)
            elif ABCPropertyGraphConstants.FIELD_POOL in v.keys():
                # pool reference
                format = DelegationFormat.PoolReference
                pool_id = v[ABCPropertyGraphConstants.FIELD_POOL]
                caporlab = None
            else:
                raise DelegationException(msg='Invalid dictionary format when parsing delegation from JSON')
            d = Delegation(atype=atype, delegation_id=k, aformat=format, pool_id=pool_id)
            if caporlab is not None:
                d.set_details(caporlab)
            ds.add_delegations(d)
        return ds

    def __repr__(self):
        return f"{self.type} consisting of {self.delegations}"


class Pool:
    """
    Label and capacity pools. Each pool knows which node defined it and which nodes it
    applies to and can be set/queried for that information. These structures help
    transform ARM into ADM, but themselves don't serialize.
    """
    def __init__(self, *, atype: DelegationType, pool_id: str, delegation_id: str = None,
                 defined_on: str = None, defined_for: List[str] = None):
        """
        Define a label or capacity pool and optionally specify where it is defined and for which nodes
        and what resources fall into it
        :param atype:
        :param pool_id:
        :param defined_on:
        :param defined_for:
        :param delegation_id:
        """
        assert atype is not None
        assert pool_id is not None
        self.type = atype
        self.on_ = defined_on
        self.delegation_id = delegation_id
        if defined_for is None:
            self.for_ = set()
        else:
            self.for_ = set(defined_for)
        if defined_on in self.for_:
            self.for_.remove(defined_on)
        self.pool_id = pool_id
        self.pool_details = None

    def set_defined_on(self, node_id: str) -> None:
        """
        Set ID of node on which this pool is defined
        :param node_id:
        :return:
        """
        assert node_id is not None
        self.on_ = node_id

    def set_defined_for(self, node_id_list: List[str]) -> None:
        """
        set/replace the pool's defined_for list of nodes
        :param node_id_list:
        :return:
        """
        assert node_id_list is not None
        assert len(node_id_list) != 0

        self.for_ = set(node_id_list)

    def add_defined_for(self, node_ids: str or List) -> None:
        """
        Add to the list of nodes for which the pool is defined for.
        Individual items or lists
        :param node_ids:
        :return:
        """
        assert node_ids is not None
        if isinstance(node_ids, str):
            self.for_.add(node_ids)
        elif isinstance(node_ids, list):
            self.for_.update(node_ids)

    def set_pool_details(self, caporlab: Capacities or Labels) ->None:
        """
        set details of the pool to a Capacities or Labels object
        :param caporlab:
        :return:
        """
        assert (isinstance(caporlab, Labels) or isinstance(caporlab, Capacities))
        self.pool_details = caporlab

    def get_pool_details(self) -> Capacities or Labels:
        """
        Return label or capacity pool details as Capacities or Labels object
        :return:
        """
        return self.pool_details

    def set_delegation_id(self, *, delegation_id: str) -> None:
        assert delegation_id is not None
        self.delegation_id = delegation_id

    def is_defined_on(self, node_id: str) -> bool:
        """
        Is this pool defined on this node?
        :param node_id:
        :return:
        """
        return node_id == self.on_

    def is_defined_for(self, node_id: str) -> bool:
        """
        Is this pool defined for this node?
        :param node_id:
        :return:
        """
        return node_id in self.for_

    def get_defined_on(self) -> str:
        return self.on_

    def get_defined_for(self) -> Set[str]:
        return self.for_

    def get_pool_id(self) -> str:
        return self.pool_id

    def get_pool_type(self) -> DelegationType:
        return self.type

    def get_delegation_id(self) -> str:
        return self.delegation_id

    def validate_pool(self):
        """
        Validate a pool to make sure delegation id is present, on, for and details are defined
        :return:
        """
        if self.delegation_id is None:
            raise PoolException(f"Pool {self.pool_id} does not have a delegation ID")
        if self.get_defined_on() is None:
            raise PoolException(f"Pool {self.pool_id} is not defined on any node")
        if self.get_defined_for() is None or len(self.get_defined_for()) == 0:
            raise PoolException(f"Pool {self.pool_id} is not mentioned on any nodes")
        if self.get_pool_details() is None:
            raise PoolException(f"Pool {self.pool_id} does not have any resource details")

    def __repr__(self) -> str:
        return f"{self.type} pool {self.pool_id} delegated to {self.delegation_id}: " \
               f"{self.on_}=> {self.for_} with {self.pool_details} "


class Pools:
    """
    Define/extract multiple pools for a given ARM, ADM or CBM model
    """
    def __init__(self, atype: DelegationType):
        self.pool_by_id = {}
        self.pools_by_delegation = None
        self.pool_type = atype

    def get_type(self) -> DelegationType:
        return self.pool_type

    def get_pool_by_id(self, *, pool_id: str, strict: bool = False) -> Pool:
        """
        Return a pool for this id. If strict is false (default) and pool doesn't exist,
        create a new pool of type matching this pools type.
        :param pool_id:
        :param strict
        :return:
        """
        assert pool_id is not None
        if strict:
            return self.pool_by_id.get(pool_id, None)
        else:
            p = self.pool_by_id.get(pool_id, None)
            if p is None:
                p = Pool(atype=self.pool_type, pool_id=pool_id)
                self.add_pool(pool=p)
            return p

    def build_index_by_delegation_id(self) -> None:
        """
        Validate all pools then index them by delegation ids, raise PoolException if any of them
        don't have the delegation id defined. Validate that all fields are filled in.
        :return:
        """
        self.pools_by_delegation = {}
        for pool in self.pool_by_id.values():
            pool.validate_pool()
            el = self.pools_by_delegation.get(pool.get_delegation_id(), list())
            el.append(pool)
            self.pools_by_delegation[pool.get_delegation_id()] = el

    def get_pools_by_delegation_id(self, delegation_id: str) -> List[Pool]:
        """
        return a list of Pool(s) for a delegation. Raises PoolException if the index based on
        delegation ids have not been built yet
        :param delegation_id:
        :return:
        """
        assert delegation_id is not None
        if self.pools_by_delegation is None:
            raise PoolException("Pools are node indexed by delegation, please run build_index_by_delegation_id()")
        return self.pools_by_delegation.get(delegation_id, None)

    def add_pool(self, *, pool: Pool) -> None:
        """
        add a pool to the index
        :param pool:
        :return:
        """
        assert pool is not None
        assert pool.get_pool_id() is not None
        if pool.get_pool_type() != self.pool_type:
            raise PoolException(msg=f'Pool type {pool.get_pool_type()} does not match Pools '
                                    f'container type {self.pool_type}')
        self.pool_by_id[pool.get_pool_id()] = pool

    def incorporate_delegation(self, *, node_id: str, deleg: Delegations):
        """
        Take a Delegations object (e.g. extracted from a model element) and
        add its information to construct pools. Ignores single element pools.
        Node id is the id of the node where it was found.
        :param node_id:
        :param deleg:
        :return:
        """
        assert node_id is not None
        assert deleg is not None
        if deleg.type != self.pool_type:
            raise PoolException(msg=f'Delegation type {deleg.type} does not match '
                                    f'Pools container type {self.pool_type}')
        # ignore single element pools in delegations
        for k, d in deleg.delegations.items():
            if d.get_format() == DelegationFormat.SinglePool:
                continue
            p = self.get_pool_by_id(pool_id=d.get_pool_name())
            if d.get_format() == DelegationFormat.PoolDefinition:
                if p.get_defined_on() is not None:
                    raise PoolException(msg=f'Pool {d.get_pool_name} has already been defined')
                p.set_defined_on(node_id)
                p.set_pool_details(d.get_details())
                p.set_delegation_id(delegation_id=d.get_delegation_id())
            else:
                # PoolReference
                p.add_defined_for(node_id)
                p.set_delegation_id(delegation_id=d.get_delegation_id())

    def generate_delegations_by_node_id(self) -> Dict[str, Delegations]:
        """
        Produce a dictionary of delegations objects indexed
        by node id they pertain to (to support serialization onto the model).
        :return:
        """
        ret = dict()
        if self.pools_by_delegation is None:
            return ret
        for delegation_id, pool_list in self.pools_by_delegation.items():
            for pool in pool_list:
                # pool definition
                pd = Delegation(atype=self.pool_type, delegation_id=delegation_id,
                                aformat=DelegationFormat.PoolDefinition, pool_id=pool.get_pool_id())
                pd.set_details(pool.get_pool_details())
                node = pool.get_defined_on()
                ds = ret.get(node, None)
                if ds is None:
                    ret[node] = Delegations(atype=self.pool_type)
                    ds = ret[node]
                ds.add_delegations(pd)
                # pool references
                for node in pool.get_defined_for():
                    pr = Delegation(atype=self.pool_type, delegation_id=delegation_id,
                                    aformat=DelegationFormat.PoolReference, pool_id=pool.get_pool_id())
                    ds = ret.get(node, None)
                    if ds is None:
                        ret[node] = Delegations(atype=self.pool_type)
                        ds = ret[node]
                    ds.add_delegations(pr)
        return ret

    def get_delegation_ids(self) -> Set:
        """
        return a set of all delegation ids
        :return:
        """
        if self.pools_by_delegation is None:
            raise PoolException("Pools are not indexed by delegation, unable to get set of delegation ids")
        return set(self.pools_by_delegation.keys())

    def get_node_ids(self, delegation_id: str) -> Set:
        """
        Get a set of nodes for a given delegation across all pools
        :param delegation_id:
        :return:
        """
        assert delegation_id is not None
        if self.pools_by_delegation is None:
            raise PoolException("Pools are not indexed by delegation, unable to get node ids by delegation id")
        ret = set()
        if delegation_id not in self.pools_by_delegation.keys():
            return ret
        for pool in self.pools_by_delegation[delegation_id]:
            ret.update(pool.get_defined_for())
        return ret

    def validate_pools(self):
        """
        Make sure all pools have a defined on and for nodes.
        :return:
        """
        for _, pool in self.pool_by_id.items():
            pool.validate_pool()

    def __repr__(self):
        return f"{self.pool_by_id}"


class PoolException(Exception):
    """
    Exception with a pool
    """
    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"Pool exception: {msg}")


class DelegationException(Exception):
    """
    Exception with a pool
    """
    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"Delegation exception: {msg}")


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
from typing import List
import networkx_query as nxq

from .abc_asm import ABCASMPropertyGraph
from ..abc_property_graph import ABCPropertyGraph, PropertyGraphQueryException
from ..networkx_property_graph import NetworkXPropertyGraph, NetworkXGraphImporter


class NetworkxASM(ABCASMPropertyGraph, NetworkXPropertyGraph):
    """
    Class implementing Abstract Slice Model on top of NetworkX
    """
    def __init__(self, *, graph_id: str, importer: NetworkXGraphImporter, logger=None):
        super().__init__(graph_id=graph_id, importer=importer, logger=logger)

    def find_node_by_name(self, node_name: str, label: str) -> str:
        """
        Return a node id of a node with this name
        :param node_name:
        :param label: label/class of the node
        :return:
        """
        assert node_name is not None
        my_graph = self.storage.get_graph(self.graph_id)
        graph_nodes = list(nxq.search_nodes(my_graph,
                                            {'and': [
                                                {'eq': [ABCPropertyGraph.GRAPH_ID, self.graph_id]},
                                                {'eq': [ABCPropertyGraph.PROP_NAME, node_name]},
                                                {'eq': [ABCPropertyGraph.PROP_CLASS, label]}
                                            ]}))
        if len(graph_nodes) == 0:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                              msg=f"Unable to find node with name {node_name} class {label}")
        if len(graph_nodes) > 1:
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=None,
                                              msg=f"Graph contains multiple nodes with name {node_name} class {label}")

        return my_graph.nodes[graph_nodes[0]][ABCPropertyGraph.NODE_ID]

    def check_node_unique(self, *, label: str, name: str):
        """
        Check no other node of this class/label and name exists
        :param label:
        :param name:
        :return:
        """
        graph_nodes = list(nxq.search_nodes(self.storage.get_graph(self.graph_id),
                                            {'and': [
                                                {'eq': [ABCPropertyGraph.GRAPH_ID, self.graph_id]},
                                                {'eq': [ABCPropertyGraph.PROP_NAME, name]},
                                                {'eq': [ABCPropertyGraph.PROP_CLASS, label]}
                                            ]}))
        return len(graph_nodes) == 0

    def check_node_name(self, *, node_id: str, label: str, name: str) -> bool:
        assert node_id is not None
        assert name is not None
        assert label is not None

        graph_nodes = list(nxq.search_nodes(self.storage.get_graph(self.graph_id),
                                            {'and': [
                                                {'eq': [ABCPropertyGraph.GRAPH_ID, self.graph_id]},
                                                {'eq': [ABCPropertyGraph.PROP_NAME, name]},
                                                {'eq': [ABCPropertyGraph.PROP_CLASS, label]},
                                                {'eq': [ABCPropertyGraph.NODE_ID, node_id]}
                                            ]}))
        return len(graph_nodes) > 0


class NetworkXASMFactory:
    """
    Help convert graphs between formats so long as they are rooted in NetworkXPropertyGraph
    """
    @staticmethod
    def create(graph: NetworkXPropertyGraph) -> NetworkxASM:
        assert graph is not None
        assert isinstance(graph.importer, NetworkXGraphImporter)

        return NetworkxASM(graph_id=graph.graph_id,
                           importer=graph.importer,
                           logger=graph.log)



from collections.abc import Mapping, Iterator


class ViewOnlyDict(Mapping):
    """
    Provides a view-only access to a dictionary
    """

    def __init__(self, d):
        self.__dict = d

    def __iter__(self) -> Iterator:
        return self.__dict.__iter__()

    def __len__(self) -> int:
        return self.__dict.__len__()

    def __getitem__(self, k):
        return self.__dict.__getitem__(k)

    def __repr__(self):
        return self.__dict.__repr__()

    def __str__(self):
        return self.__dict.__str__()

    def __hash__(self):
        return self.__dict.__hash__()


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
from typing import Tuple, Any, List

import uuid

from .model_element import ModelElement, ElementType

from fim.graph.abc_property_graph import ABCPropertyGraph
from fim.user.interface import Interface
from fim.slivers.network_link import NetworkLinkSliver, LinkType
from fim.slivers.switch_fabric import SFLayer


class Link(ModelElement):
    """
    A link object in a topology. In addition to public methods the following calls
    return various dictionaries or lists:
    link.interface_list - a list of interfaces attached to this link
    """
    def __init__(self, *, name: str, node_id: str = None, topo: Any,
                 etype: ElementType = ElementType.EXISTING,
                 interfaces: List[Interface] = None,
                 layer: SFLayer = SFLayer.L2,
                 ltype: LinkType = None, technology: str = None, **kwargs):
        """
        Don't call this method yourself, call topology.add_link()
        node_id will be generated if not provided for experiment topologies

        :param name:
        :param node_id:
        :param topo:
        :param etype: is this supposed to exist or new should be created
        :param interfaces: list of interface objects to connect
        :param layer: layer of the link, defaults to L2
        :param ltype: link type if new
        :param technology: link technology
        :param kwargs: any additional properties
        """
        assert name is not None
        assert topo is not None

        if etype == ElementType.NEW:
            # cant use isinstance as it would create circular import dependencies
            # node id myst be specified for new nodes in substrate topologies
            if str(topo.__class__) == "<class 'fim.user.topology.SubstrateTopology'>" and \
                    node_id is None:
                raise RuntimeError("When adding new links to substrate topology nodes you must specify static Node ID")
            if node_id is None:
                node_id = str(uuid.uuid4())
            super().__init__(name=name, node_id=node_id, topo=topo)
            if ltype is None:
                raise RuntimeError("When creating new links you must specify LinkType")
            # FIXME isinstance
            if interfaces is None or len(interfaces) == 0 or (not isinstance(interfaces, tuple) and
                                                              not isinstance(interfaces, list)):
                raise RuntimeError("When creating new links you must specify the list of interfaces to connect")

            self._interfaces = interfaces
            sliver = NetworkLinkSliver()
            sliver.node_id = self.node_id
            sliver.set_name(self.name)
            sliver.set_type(ltype)
            sliver.set_layer(layer)
            sliver.set_technology(technology)
            sliver.set_properties(**kwargs)

            # get a list of node_ids for interfaces
            interface_ids = (iff.node_id for iff in interfaces)
            self.topo.graph_model.add_network_link_sliver(interfaces=interface_ids, lsliver=sliver)
        else:
            assert node_id is not None
            super().__init__(name=name, node_id=node_id, topo=topo)
            # check that this node exists
            existing_node_id = self.topo.\
                graph_model.find_node_by_name(node_name=name,
                                              label=ABCPropertyGraph.CLASS_Link)
            if existing_node_id != node_id:
                raise RuntimeError(f'Link name {name} is not unique within the topology.')
            # collect a list of interfaces it attaches to
            interface_list = self.topo.graph_model.get_all_link_interfaces(link_id=self.node_id)
            name_id_tuples = list()
            # need to look up their names - a bit inefficient, need to think about this /ib
            for iff in interface_list:
                _, props = self.topo.graph_model.get_node_properties(node_id=iff)
                name_id_tuples.append((props[ABCPropertyGraph.PROP_NAME], iff))
            self._interfaces = [Interface(node_id=tup[1], topo=topo, name=tup[0]) for tup in name_id_tuples]

    def get_property(self, pname: str) -> Any:
        """
        Retrieve a interface property
        :param pname:
        :return:
        """
        _, node_properties = self.topo.graph_model.get_node_properties(node_id=self.node_id)
        if_sliver = self.topo.graph_model.link_sliver_from_graph_properties_dict(node_properties)
        return if_sliver.get_property(pname)

    def set_property(self, pname: str, pval: Any):
        """
        Set a interface property
        :param pname:
        :param pval:
        :return:
        """
        link_sliver = NetworkLinkSliver()
        link_sliver.set_property(prop_name=pname, prop_val=pval)
        # write into the graph
        prop_dict = self.topo.graph_model.link_sliver_to_graph_properties_dict(link_sliver)
        self.topo.graph_model.update_node_properties(node_id=self.node_id, props=prop_dict)

    def set_properties(self, **kwargs):
        """
        Set multiple properties of the link
        :param kwargs:
        :return:
        """
        link_sliver = NetworkLinkSliver()
        link_sliver.set_properties(**kwargs)
        # write into the graph
        prop_dict = self.topo.graph_model.link_sliver_to_graph_properties_dict(link_sliver)
        self.topo.graph_model.update_node_properties(node_id=self.node_id, props=prop_dict)

    @staticmethod
    def list_properties() -> Tuple[str]:
        return NetworkLinkSliver.list_properties()

    def __list_of_interfaces(self) -> List[Interface] or None:
        """
        Make a copy and return a list of interface objects
        :return:
        """
        if self._interfaces is not None:
            return self._interfaces.copy()
        return None

    def __getattr__(self, item):
        """
        Special handling for attributes like 'nodes' and 'links' -
        which query into the model. They return dicts and list
        containers. Modifying containers does not affect the underlying
        graph mode, but modifying elements of lists or values of dicts does.
        :param item:
        :return:
        """

        if item == 'interface_list':
            return self.__list_of_interfaces()
        raise RuntimeError(f'Attribute {item} not available')

    def __repr__(self):
        _, node_properties = self.topo.graph_model.get_node_properties(node_id=self.node_id)
        link_sliver = self.topo.graph_model.link_sliver_from_graph_properties_dict(node_properties)
        interface_names = [iff.name for iff in self._interfaces]
        return link_sliver.__repr__() + str(interface_names)

    def __str__(self):
        return self.__repr__()


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
from typing import List, Dict
import json

from abc import ABC, abstractmethod

from fim.graph.abc_property_graph_constants import ABCPropertyGraphConstants


class JSONField(ABC):

    @abstractmethod
    def set_fields(self, **kwargs):
        """
        Set the fields of this object
        :param kwargs:
        :return:
        """

    def to_json(self) -> str:
        """
        Dumps to JSON the __dict__ of the instance. Be careful as the fields in this
        class should only be those that can be present in JSON output.
        If there are no values in the object, returns empty string.
        :return:
        """
        d = self.__dict__.copy()
        for k in self.__dict__:
            if d[k] is None or d[k] == 0:
                d.pop(k)
        if len(d) == 0:
            return ''
        return json.dumps(d, skipkeys=True, sort_keys=True)

    @classmethod
    def from_json(cls, json_string: str):
        """
        Set fields from json string and returns a new object
        :param json_string:
        :return: object
        """
        if json_string is None or len(json_string) == 0 or json_string == ABCPropertyGraphConstants.NEO4j_NONE:
            return None
        d = json.loads(json_string)
        ret = cls()
        ret.set_fields(**d)
        return ret

    def to_dict(self) -> Dict[str, str] or None:
        """
        Convert to a dictionary skipping empty fields. Returns None
        if all fields are empty
        :return:
        """
        d = self.__dict__.copy()
        for k in self.__dict__:
            if d[k] is None or d[k] == 0:
                d.pop(k)
        if len(d) == 0:
            return None
        return d

    def __repr__(self):
        return self.to_json()

    def __str__(self):
        return self.to_json()

    def list_fields(self):
        l = list(self.__dict__.keys())
        l.sort()
        return l


class Capacities(JSONField):
    """
    Implements basic capacity field handling - encoding and decoding
    from JSON dictionaries of properties
    """
    UNITS = {'cpu': '', 'unit': '', 'core': '', 'ram': 'G', 'disk': 'G', 'bw': 'Gbps'}

    def __init__(self):
        self.cpu = 0
        self.core = 0
        self.ram = 0
        self.disk = 0
        self.bw = 0
        self.unit = 0

    def set_fields(self, **kwargs):
        """
        Universal integer setter for all fields.
        Values should be non-negative integers. Throws a CapacityException
        if you try to set a non-existent field.
        :param kwargs:
        :return: self to support call chaining
        """
        for k, v in kwargs.items():
            if v is not None:
                assert v >= 0
            try:
                # will toss an exception if field is not defined
                self.__getattribute__(k)
                self.__setattr__(k, v)
            except AttributeError:
                raise CapacityException(f"Unable to set field {k} of capacity, no such field available")
        return self

    def __add__(self, other):
        assert isinstance(other, Capacities)
        ret = Capacities()
        for f, v in self.__dict__.items():
            ret.__dict__[f] = self.__dict__[f] + other.__dict__[f]

        return ret

    def __sub__(self, other):
        assert isinstance(other, Capacities)
        ret = Capacities()

        for f, v in self.__dict__.items():
            ret.__dict__[f] = self.__dict__[f] - other.__dict__[f]

        return ret

    def negative_fields(self) -> List[str]:
        """
        returns list of fields that are negative
        :return:
        """
        ret = list()
        for f, v in self.__dict__.items():
            if v < 0:
                ret.append(f)

        return ret

    def positive_fields(self, fields: List[str]) -> bool:
        """
        Return true if indicated fields are positive >0
        :param fields:
        :return:
        """
        assert fields is not None
        for f in fields:
            if self.__dict__[f] <= 0:
                return False
        return True

    def __str__(self):
        d = self.__dict__.copy()
        for k in self.__dict__:
            if d[k] is None or d[k] == 0:
                d.pop(k)
        if len(d) == 0:
            return ''
        ret = "{ "
        for i, v in d.items():
            ret = ret + i + ": " + f'{v:,} ' + self.UNITS[i] + ", "
        return ret[:-2] + "}"


class CapacityTuple:
    """
    This class takes two capacities objects (what is the total
    available and what is free) and helps print them.
    """
    def __init__(self, *, total: Capacities, allocated: Capacities):
        assert total is not None
        if allocated is None:
            allocated = Capacities()
        self.available = total
        self.free = total - allocated

    def __str__(self):
        d2 = self.available.__dict__.copy()
        d1 = self.free.__dict__.copy()

        for k in self.free.__dict__:
            if d1[k] == 0 and d2[k] == 0:
                d1.pop(k)
                d2.pop(k)
        if len(d1) == 0:
            return ''
        ret = '{ '
        for k in d1:
            ret = ret + k + ": " + f'{d1[k]:,}' + "/" + f'{d2[k]:,} ' + Capacities.UNITS[k] + ", "
        return ret[:-2] + '}'


class Labels(JSONField):
    """
    Class implementing various encodings of labels field, encoding
    and decoding from JSON dictionaries of properties
    """
    def __init__(self):
        self.bdf = None
        self.mac = None
        self.ipv4 = None
        self.ipv4_range = None
        self.ipv6 = None
        self.ipv6_range = None
        self.asn = None
        self.vlan = None
        self.vlan_range = None
        self.instance = None
        self.instance_parent = None
        self.local_name = None

    def set_fields(self, **kwargs):
        """
        Universal setter for all fields. Values should be strings or lists of strings.
        Throws a LabelException if you try to set a non-existent field.
        :param kwargs:
        :return: self to support call chaining
        """
        for k, v in kwargs.items():
            assert v is not None  # could be strings or lists of strings
            assert isinstance(v, str) or isinstance(v, list)
            try:
                # will toss an exception if field is not defined
                self.__getattribute__(k)
                self.__setattr__(k, v)
            except AttributeError:
                raise LabelException(f"Unable to set field {k} of labels, no such field available")
        # to support call chaining
        return self

    def __add__(self, other):
        assert isinstance(other, Labels)
        # FIXME: does ADD mean finding a union of label sets per type?
        raise RuntimeError("Not Implemented")

    def __str__(self):
        d = self.__dict__.copy()
        for k in self.__dict__:
            if d[k] is None:
                d.pop(k)
        if len(d) == 0:
            return ''
        ret = "{ "
        for i, v in d.items():
            ret = ret + i + ": " + str(v) + ", "
        return ret[:-2] + "}"


class ReservationInfo(JSONField):
    """
    Reservation info structure for ASM sliver objects
    """

    def __init__(self):
        self.reservation_id = None
        self.reservation_state = None

    def set_fields(self, **kwargs):
        """
        Universal setter for all fields. Values should be strings or lists of strings.
        Throws a ReservationInfoException if you try to set a non-existent field.
        :param kwargs:
        :return: self to support call chaining
        """
        for k, v in kwargs.items():
            assert v is not None  # could be strings or lists of strings
            assert isinstance(v, str) or isinstance(v, list)
            try:
                # will toss an exception if field is not defined
                self.__getattribute__(k)
                self.__setattr__(k, v)
            except AttributeError:
                raise ReservationInfoException(f"Unable to set field {k} of reservation info, no such field available")
        # to support call chaining
        return self


class StructuralInfo(JSONField):
    """
    Structural info on the for sliver objects (things like - is it a stitching node,
    what is parent graph or subgraph)
    """
    def __init__(self):
        self.sub_graph_id = None
        self.parent_graph_id = None
        self.adm_graph_ids = None

    def set_fields(self, **kwargs):
        """
        Universal setter for all fields. Values are strings or boolean.
        Throws a
        :param kwargs:
        :return:
        """
        for k, v in kwargs.items():
            assert v is not None  # could be strings or lists of strings
            assert isinstance(v, str) or isinstance(v, list)
            try:
                # will toss an exception if field is not defined
                self.__getattribute__(k)
                self.__setattr__(k, v)
            except AttributeError:
                raise StructuralInfoException(f"Unable to set field {k} of structural info, no such field available")
        # to support call chaining
        return self


class CapacityException(Exception):
    """
    Exception with a capacity
    """
    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"Delegation exception: {msg}")


class LabelException(Exception):
    """
    Exception with a label
    """
    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"Label exception: {msg}")


class ReservationInfoException(Exception):
    """
    Exception with a reservation info
    """
    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"ReservationInfo exception: {msg}")


class StructuralInfoException(Exception):
    """
    Exception with a structural info
    """

    def __init__(self, msg: str):
        assert msg is not None
        super().__init__(f"ReservationInfo exception: {msg}")


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)

from typing import Any, Dict, List, Tuple

import uuid

from fim.view_only_dict import ViewOnlyDict
from ..graph.abc_property_graph import ABCPropertyGraph
from .model_element import ModelElement, ElementType
from .interface import Interface
from .switch_fabric import SwitchFabric, SFLayer
from ..slivers.capacities_labels import Labels
from ..slivers.attached_components import ComponentSliver, ComponentType
from ..slivers.component_catalog import ComponentCatalog


class Component(ModelElement):
    """
    A component, like a GPU, a NIC, an FPGA or an NVMe drive of a node in a model.
    In addition to public methods the following calls
    return various dictionaries or lists:
    component.interfaces - a dictionary of interfaces
    component.interface_list - a list of interfaces
    component.switch_fabrics - a dictionary of switch fabrics
    """

    def __init__(self, *, name: str, node_id: str = None, topo: Any,
                 etype: ElementType = ElementType.EXISTING, model: str = None,
                 ctype: ComponentType = None, parent_node_id: str = None,
                 switch_fabric_node_id: str = None, interface_node_ids: List[str] = None,
                 interface_labels: List[Labels] = None, **kwargs):
        """
        Don't call this yourself, use Node.add_component(). Instantiates components based on
        catalog resource file.
        :param name:
        :param node_id:
        :param topo:
        :param etype: is this supposed to be new or existing
        :param model: must be specified if a new component
        :param ctype: component type
        :param parent_node_id: node_id of the parent Node (for new components)
        :param switch_fabric_node_id: node id of switch fabric if one needs to be added (for substrate models only)
        :param interface_node_ids: a list of node ids for expected interfaces (for substrate models only)
        :param interface_labels: a list of Labels structure to associate with each interface
        """

        assert name is not None
        assert topo is not None

        if etype == ElementType.NEW:
            # cant use isinstance as it would create circular import dependencies
            if str(topo.__class__) == "<class 'fim.user.topology.SubstrateTopology'>" and node_id is None:
                raise RuntimeError("When adding components to substrate topology nodes you must specify static Node ID")
            if node_id is None:
                node_id = str(uuid.uuid4())
            if parent_node_id is None:
                raise RuntimeError("Parent node id must be specified for new components")
            if model is None or ctype is None:
                raise RuntimeError("Model and component type must be specified for new components")
            if str(topo.__class__) == "<class 'fim.user.topology.SubstrateTopology'>" and \
                    (ctype == ComponentType.SharedNIC or ctype == ComponentType.SmartNIC) and \
                    (switch_fabric_node_id is None or interface_node_ids is None or interface_labels is None):
                raise RuntimeError('For substrate topologies and components with network interfaces '
                                   'static switch_fabric node id, interface node ids and interface labels'
                                   'must be specified')
            super().__init__(name=name, node_id=node_id, topo=topo)
            cata = ComponentCatalog()

            comp_sliver = cata.generate_component(name=name, model=model, ctype=ctype,
                                                  switch_fabric_node_id=switch_fabric_node_id,
                                                  interface_node_ids=interface_node_ids,
                                                  interface_labels=interface_labels)
            comp_sliver.node_id = node_id
            comp_sliver.set_properties(**kwargs)

            self.topo.graph_model.add_component_sliver(parent_node_id=parent_node_id, component=comp_sliver)
        else:
            assert node_id is not None
            super().__init__(name=name, node_id=node_id, topo=topo)
            if not self.topo.graph_model.check_node_name(node_id=node_id, name=name,
                                                         label=ABCPropertyGraph.CLASS_Component):
                raise RuntimeError(f"Component with this id and name {name} doesn't exist")

    def get_property(self, pname: str) -> Any:
        """
        Retrieve a component property
        :param pname:
        :return:
        """
        _, node_properties = self.topo.graph_model.get_node_properties(node_id=self.node_id)
        comp_sliver = self.topo.graph_model.component_sliver_from_graph_properties_dict(node_properties)
        return comp_sliver.get_property(pname)

    def set_property(self, pname: str, pval: Any):
        """
        Set a component property
        :param pname:
        :param pval:
        :return:
        """
        comp_sliver = ComponentSliver()
        comp_sliver.set_property(prop_name=pname, prop_val=pval)
        # write into the graph
        prop_dict = self.topo.graph_model.component_sliver_to_graph_properties_dict(comp_sliver)
        self.topo.graph_model.update_node_properties(node_id=self.node_id, props=prop_dict)

    def set_properties(self, **kwargs):
        """
        Set multiple properties of the component
        :param kwargs:
        :return:
        """
        comp_sliver = ComponentSliver()
        comp_sliver.set_properties(**kwargs)
        # write into the graph
        prop_dict = self.topo.graph_model.component_sliver_to_graph_properties_dict(comp_sliver)
        self.topo.graph_model.update_node_properties(node_id=self.node_id, props=prop_dict)

    @staticmethod
    def list_properties() -> Tuple[str]:
        return ComponentSliver.list_properties()

    def __get_interface_by_id(self, node_id: str) -> Interface:
        """
        Get an interface of a node by its node_id, return Interface object
        :param node_id:
        :return:
        """
        assert node_id is not None
        _, node_props = self.topo.graph_model.get_node_properties(node_id=node_id)
        assert node_props.get(ABCPropertyGraph.PROP_NAME, None) is not None
        return Interface(name=node_props[ABCPropertyGraph.PROP_NAME], node_id=node_id,
                         topo=self.topo)

    def __get_sf_by_id(self, node_id: str) -> SwitchFabric:
        """
        Get an switch fabric of a node by its node_id, return SwitchFabric object
        :param node_id:
        :return:
        """
        assert node_id is not None
        _, node_props = self.topo.graph_model.get_node_properties(node_id=node_id)
        assert node_props.get(ABCPropertyGraph.PROP_NAME, None) is not None
        return SwitchFabric(name=node_props[ABCPropertyGraph.PROP_NAME], node_id=node_id,
                            topo=self.topo)

    def __list_interfaces(self) -> ViewOnlyDict:
        """
        List all interfaces of the component as a dictionary
        :return:
        """
        node_id_list = self.topo.graph_model.get_all_node_or_component_connection_points(parent_node_id=self.node_id)
        # Could consider using frozendict here
        ret = dict()
        for nid in node_id_list:
            i = self.__get_interface_by_id(nid)
            ret[i.name] = i
        return ViewOnlyDict(ret)

    def __list_of_interfaces(self) -> Tuple[Interface]:
        """
        List all interfaces of the component
        :return:
        """
        node_id_list = self.topo.graph_model.get_all_node_or_component_connection_points(parent_node_id=self.node_id)
        # Could consider using frozendict here
        ret = list()
        for nid in node_id_list:
            i = self.__get_interface_by_id(nid)
            ret.append(i)
        return tuple(ret)

    def __list_switch_fabrics(self) -> ViewOnlyDict:
        """
        List all switch fabric children of a node as a dictionary organized
        by switch fabric name. Modifying the dictionary will not affect
        the underlying model, but modifying Components in the dictionary will.
        :return:
        """
        node_id_list = self.topo.graph_model.get_all_network_node_or_component_sfs(parent_node_id=self.node_id)
        # Could consider using frozendict or other immutable idioms
        ret = dict()
        for nid in node_id_list:
            c = self.__get_sf_by_id(nid)
            ret[c.name] = c
        return ViewOnlyDict(ret)

    def __getattr__(self, item):
        """
        Special handling for attributes like 'components' and 'interfaces' -
        which query into the model. They return dicts and list
        containers. Modifying containers does not affect the underlying
        graph mode, but modifying elements of lists or values of dicts does.
        :param item:
        :return:
        """
        if item == 'interfaces':
            return self.__list_interfaces()
        if item == 'interface_list':
            return self.__list_of_interfaces()
        if item == 'switch_fabrics':
            return self.__list_switch_fabrics()
        raise RuntimeError(f'Attribute {item} not available')

    def __repr__(self):
        _, node_properties = self.topo.graph_model.get_node_properties(node_id=self.node_id)
        comp_sliver = self.topo.graph_model.component_sliver_from_graph_properties_dict(node_properties)
        return comp_sliver.__repr__()

    def __str__(self):
        return self.__repr__()




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
"""
Neo4j implementation of CBM (Combined Broker Model) functionality.
"""
import uuid
import json

from typing import List, Dict, Tuple

from collections import defaultdict

from ..abc_property_graph import ABCPropertyGraph, ABCPropertyGraphConstants, PropertyGraphQueryException
from ..neo4j_property_graph import Neo4jPropertyGraph, Neo4jGraphImporter
from .abc_cbm import ABCCBMPropertyGraph
from .neo4j_adm import Neo4jADMGraph
from fim.slivers.capacities_labels import StructuralInfo, StructuralInfoException
from fim.slivers.delegations import DelegationType, Delegations
from fim.slivers.attached_components import AttachedComponentsInfo

from ...pluggable import PluggableRegistry, BrokerPluggable, PluggableType


class Neo4jCBMGraph(Neo4jPropertyGraph, ABCCBMPropertyGraph):
    """
    Neo4j implementation of CBM
    """
    BQM_MERGED_FIELDS = ['LabelDelegations', 'CapacityDelegations']
    DELEGATION_TYPE_TO_PROP_NAME = {DelegationType.LABEL: ABCPropertyGraph.PROP_LABEL_DELEGATIONS,
                                    DelegationType.CAPACITY: ABCPropertyGraph.PROP_CAPACITY_DELEGATIONS
                                    }

    def __init__(self, *, graph_id: str = None, importer: Neo4jGraphImporter, logger=None):
        """
        Initialize CBM either from existing ID or generate new graph id
        :param graph_id:
        :param importer:
        """
        if graph_id is None:
            graph_id = str(uuid.uuid4())

        super().__init__(graph_id=graph_id, importer=importer, logger=logger)

    def _update_node_delegations(self, *, node_id: str, adm: ABCPropertyGraph) -> None:
        """
        Update the CBM node by-ADM-id delegation dictionary (it is assumed to exist already
        on common nodes) from corresponding ADM node.
        Must be invoked before ADM is merged into CBM.
        :param node_id:
        :param adm:
        :return:
        """
        _, cbm_node_props = self.get_node_properties(node_id=node_id)
        _, adm_node_props = adm.get_node_properties(node_id=node_id)
        props_modified = False
        for delegation_prop_name in [self.PROP_LABEL_DELEGATIONS, self.PROP_CAPACITY_DELEGATIONS]:
            # one or both of properties can have data, however only CBM or ADM
            # can have information about delegations - only one aggregate can speak for a resource
            if cbm_node_props.get(delegation_prop_name, None) is None:
                cbm_delegations = None
            else:
                cbm_delegations = Delegations.from_json(json_str=cbm_node_props[delegation_prop_name],
                                                        atype=DelegationType.LABEL if
                                                        delegation_prop_name == self.PROP_LABEL_DELEGATIONS else
                                                        DelegationType.CAPACITY)

            if adm_node_props.get(delegation_prop_name, None) is None:
                adm_delegations = None
            else:
                adm_delegations = Delegations.from_json(json_str=adm_node_props[delegation_prop_name],
                                                        atype=DelegationType.LABEL if
                                                        delegation_prop_name == self.PROP_LABEL_DELEGATIONS else
                                                        DelegationType.CAPACITY)

            if adm_delegations is None and cbm_delegations is None:
                continue
            if adm_delegations is not None and cbm_delegations is not None:
                raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_id,
                                                  msg=f'This node contains delegations from both CBM and ADM graph,'
                                                      f'which is not allowed.')
            props_modified = True
            # take the non-None dictionary and write back
            new_delegations = cbm_delegations if cbm_delegations is not None else adm_delegations
            cbm_node_props[delegation_prop_name] = new_delegations.to_json()
        if props_modified:
            # write back
            self.update_node_properties(node_id=node_id, props=cbm_node_props)

    def merge_adm(self, *, adm: ABCPropertyGraph) -> None:
        """
        Merge an ADM into the CBM
        :param adm:
        :return:
        """
        assert adm is not None
        assert adm.graph_exists()

        # clone ADM with temporary ID - we will be changing its properties
        temp_adm_graph_id = str(uuid.uuid4())
        self.log.info('CREATED TEMPORARY ADM GRAPH ID ' + temp_adm_graph_id +
                      ' when merging graph ' + adm.graph_id + ' into ' + self.graph_id)
        temp_graph = adm.clone_graph(new_graph_id=temp_adm_graph_id)
        # crude typecasting
        temp_adm_graph = Neo4jADMGraph(graph_id=temp_graph.graph_id,
                                       importer=temp_graph.importer,
                                       logger=temp_graph.log)

        # rewrite delegations into dictionaries keyed by ADM graph id,
        # but be sure to use original ADM graph ID when rewriting
        temp_adm_graph.rewrite_delegations(real_adm_id=adm.graph_id)

        # update the structural info adm_graph_ids field to a single member list on all CBM nodes
        si = StructuralInfo().set_fields(adm_graph_ids=[adm.graph_id])
        temp_adm_graph.update_nodes_property(prop_name=self.PROP_STRUCTURAL_INFO,
                                             prop_val=si.to_json())

        # if CBM is empty, just force ADM into it
        if not self.graph_exists():
            temp_adm_graph.update_nodes_property(prop_name=ABCPropertyGraphConstants.GRAPH_ID,
                                                 prop_val=self.graph_id)
            return

        # CBM is not empty - need to merge ADM with it

        # locate nodes with matching IDs
        common_node_ids = self.find_matching_nodes(other_graph=temp_adm_graph)

        # Merging CBM properties with ADM (on merged nodes):
        # edited after the fact.
        # GraphID: use CBM
        # NodeID: use CBM (should be same anyway)
        # Labels: use CBM - may have information from reservations already
        # Capacities: use CBM - may have information from reservations already - should be same
        # in both graphs initially
        # LabelDelegations: delegations can be provided only by one graph - either CBM or ADM
        # CapacityDelegations: delegations can be provided only by one graph - either CBM or ADM
        for node_id in common_node_ids:
            # update Label and Capacity delegation prior to merge
            # when nodes are merged, these properties are kept, ADM properties
            # are discarded
            self._update_node_delegations(node_id=node_id, adm=temp_adm_graph)

            # default behavior - keep CBM node properties, discard
            # temp_adm_graph properties
            self.merge_nodes(node_id=node_id,
                             other_graph=temp_adm_graph)

            # add delegation graph id to ADMGraphIDs property on common (merged) nodes
            _, cbm_node_props = self.get_node_properties(node_id=node_id)
            si = StructuralInfo.from_json(cbm_node_props[self.PROP_STRUCTURAL_INFO])
            si.adm_graph_ids.append(adm.graph_id)
            self.update_node_property(node_id=node_id, prop_name=self.PROP_STRUCTURAL_INFO,
                                      prop_val=si.to_json())

        # rewrite GraphID on the remaining nodes of
        # temporary ADM graph (after that it ceases to exist)
        # NOTE: this takes advantage of Neo4j semantics of common store for all graphs
        # and changing the GraphID property effectively makes graph takes on a new identity
        temp_adm_graph.update_nodes_property(prop_name=ABCPropertyGraphConstants.GRAPH_ID,
                                             prop_val=self.graph_id)

    def unmerge_adm(self, *, graph_id: str) -> None:
        # Search ADMGraphIDs property and remove those nodes where it is the only
        # member of the list. For those that have two members, update the list
        # to remove.
        delete_nodes = list()
        for node in self.list_all_node_ids():
            _, adm_node_props = self.get_node_properties(node_id=node)
            si = StructuralInfo.from_json(adm_node_props[self.PROP_STRUCTURAL_INFO])

            if si.adm_graph_ids is None:
                self.log.warn(f'When unmerging ADMs encountered node {node} '
                              f'without adm_graph_ids field in property {adm_node_props[self.PROP_STRUCTURAL_INFO]}')
                continue
            if not isinstance(si.adm_graph_ids, list):
                raise PropertyGraphQueryException(node_id=node, graph_id=self.graph_id,
                                                  msg=f"When unmerging graph {graph_id}, encountered wrongly"
                                                      f"formatted {self.PROP_STRUCTURAL_INFO} field "
                                                      f"{adm_node_props[self.PROP_STRUCTURAL_INFO]}")
            if graph_id in si.adm_graph_ids:
                # update the ADM Graph IDs field
                si.adm_graph_ids.remove(graph_id)
                if len(si.adm_graph_ids) == 0:
                    delete_nodes.append(node)
                else:
                    # update
                    self.update_node_property(node_id=node, prop_name=self.PROP_STRUCTURAL_INFO,
                                              prop_val=si.to_json())
            # also unmerge delegations
            for del_prop in [self.PROP_CAPACITY_DELEGATIONS, self.PROP_LABEL_DELEGATIONS]:
                if del_prop not in adm_node_props.keys():
                    continue
                delegations = Delegations.from_json(json_str=adm_node_props[del_prop],
                                                    atype=DelegationType.LABEL if
                                                    del_prop == self.PROP_LABEL_DELEGATIONS
                                                    else DelegationType.CAPACITY)
                if delegations is None:
                    continue
                if graph_id in delegations.get_delegation_ids():
                    delegations.remove_by_id(graph_id)
                    if len(delegations.get_delegation_ids()) != 0:
                        raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node,
                                                          msg=f'This node had more than one delegation, remaining:'
                                                              f'{delegations}')
                    # under normal circumstances we should erase the delegation after unmerge if
                    # it belonged to the graph being removed
                    self.update_node_property(node_id=node, prop_name=del_prop,
                                              prop_val='')

        # remove the merged nodes
        for node in delete_nodes:
            self.delete_node(node_id=node)

    def get_bqm(self, **kwargs) -> ABCPropertyGraph:
        """
        Get a Broker Query Model (e.g. in response to a listResources() call)
        :param kwargs:
        :return:
        """
        p = PluggableRegistry()
        if p.pluggable_registered(t=PluggableType.Broker) and \
                BrokerPluggable.PLUGGABLE_PRODUCE_BQM in p.get_implemented_methods(t=PluggableType.Broker):
            # invoke the callable for plug_produce_bqm() otherwise default behavior
            c = p.get_method_callable(t=PluggableType.Broker,
                                      method=BrokerPluggable.PLUGGABLE_PRODUCE_BQM)
            return c(cbm=self, **kwargs)
        # default for now - simplest possible - return a cloned CBM

        bqm = self.clone_graph(new_graph_id=str(uuid.uuid4()))
        return bqm

    def get_delegations(self, *, node_id: str, adm_id: str, delegation_type: DelegationType) -> List or None:
        """
        Retrieve Label or Capacity delegations attribuited to a particular ADM. A delegation property
        is a dictionary of per ADM graph id delegations. Individual delegations
        are lists of dictionaries with possible pool designators. Returns None if delegation property is
        not defined at all or delegation dictionary doesn't contain ADM graph id.
        :param node_id:
        :param adm_id:
        :param delegation_type:
        :return:
        """
        assert node_id is not None
        assert adm_id is not None
        prop_field_name = self.DELEGATION_TYPE_TO_PROP_NAME[delegation_type]
        deleg_prop_val = self.get_node_json_property_as_object(node_id=node_id, prop_name=prop_field_name)
        if deleg_prop_val is None:
            return None
        if not isinstance(deleg_prop_val, dict):
            raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=node_id,
                                              msg=f"Expected CBM {delegation_type.name} "
                                                  f"delegation to be a dictionary, got {deleg_prop_val} "
                                                  f"of {type(deleg_prop_val)}")
        #print(f"Looking for {delegation_type} delegation from {adm_id} in {deleg_prop_val}")
        return deleg_prop_val.get(adm_id, None)

    def get_matching_nodes_with_components(self, *, label: str, props: Dict,
                                           comps: AttachedComponentsInfo = None) -> List[str]:
        assert label is not None
        assert props is not None

        # collect unique types, models and count them
        component_counts = defaultdict(int)
        isset = False
        if comps is not None:
            for comp in comps.list_devices():
                assert(comp.resource_model is not None or comp.resource_type is not None)
                component_counts[(comp.resource_type, comp.resource_model)] = \
                    component_counts[(comp.resource_type, comp.resource_model)] + 1
                isset = True
        # unroll properties
        node_props = ", ".join([x + ": " + '"' + props[x] + '"' for x in props.keys()])

        if not isset:
            # simple query on the properties of the node
            query = f"MATCH(n:{label} {{GraphID: $graphId, {node_props} }}) RETURN collect(n.NodeID) as candidate_ids"
        else:
            # build a query list
            node_query = f"MATCH(n:{label} {{GraphID: $graphId, {node_props} }}) WHERE "
            component_clauses = list()
            # add a clause for every tuple
            for k, v in component_counts.items():
                comp_props_list = list()
                if k[0] is not None:
                    comp_props_list.append('Type: ' + '"' + str(k[0]) + '"' + ' ')
                if k[1] is not None:
                    comp_props_list.append('Model: ' + '"' + k[1] + '"' + ' ')
                comp_props = ", ".join(comp_props_list)

                component_clauses.append(f"size( (n) -[:has]- (:Component {{GraphID: $graphId, "
                                         f"{comp_props}}}))>={str(v)} ")
            query = node_query + " and ".join(component_clauses) + " RETURN collect(n.NodeID) as candidate_ids"
        #print(f'QUERY= {query}')
        #print(f'GraphID= {self.graph_id}')
        with self.driver.session() as session:

            val = session.run(query, graphId=self.graph_id).single()
            #print(f'VAL= {val}')
        if val is None:
            return list()
        return val.data()['candidate_ids']

    def get_intersite_links(self) -> List[Tuple[str, str, str]]:
        # does a lexicographic comparison of Site properties to ensure only unique links are returned
        query = 'match p= (n:NetworkNode {Type:"Switch", GraphID: $graphId}) -[:has]- (:SwitchFabric) ' \
                '-[:connects]- (cp1:ConnectionPoint) -[:connects]- (l:Link) ' \
                '-[:connects]- (cp2:ConnectionPoint) -[:connects]- (:SwitchFabric) ' \
                '-[:has]- (m:NetworkNode {Type:"Switch", GraphID: $graphId}) where n.Site > m.Site ' \
                'return n.NodeID as source, l.NodeID as link, m.NodeID as sink, n.Site as source_site, ' \
                'm.Site as sink_site, cp1.NodeID as source_cp, cp2.NodeID as sink_cp'
        with self.driver.session() as session:
            val = session.run(query, graphId=self.graph_id).values()
        if val is None:
            return list()
        return val


class Neo4jCBMFactory:
    """
    Help convert graphs between formats so long as they are rooted in Neo4jPropertyGraph
    """
    @staticmethod
    def create(graph: Neo4jPropertyGraph) -> Neo4jCBMGraph:
        assert graph is not None
        assert isinstance(graph.importer, Neo4jGraphImporter)

        return Neo4jCBMGraph(graph_id=graph.graph_id,
                             importer=graph.importer,
                             logger=graph.log)
